#include "tensor_runtime.h"
#include "visc.h"
#include <iostream>
#include <cassert>

const size_t n_channels = 3;

Tensor *gaussianFilter_(float div) {
  std::vector<float> gauss_data = {1,  4, 6,  4,  1,  4, 16, 24, 16,
                                   4,  6, 24, 36, 24, 6, 4,  16, 24,
                                   16, 4, 1,  4,  6,  4, 1};
  for (float &f : gauss_data)
    f /= div;
  return (Tensor *)createFilterFromData(
      CUDNN_DATA_FLOAT, gauss_data.data(), 5, 5, 1);
}

Tensor *gaussianFilter() { return gaussianFilter_(16.0); }

void *normalize(size_t &id, void *image) {
  auto *max_1D = wrapper_tensorReduce(
      std::to_string(id++).c_str(), image, 2, (int)MathOp::Max);
  auto *max = wrapper_tensorReduce(
      std::to_string(id++).c_str(), max_1D, 3, (int)MathOp::Max);
  auto *img_norm = wrapper_tensorMap2(
      std::to_string(id++).c_str(), (int)MathOp::Div, image, max);
  freeTensor(max_1D);
  freeTensor(max);
  return img_norm;
}

extern std::vector<size_t> sizes(Tensor *t);

void forward_reshape(void *t) {
  auto *tensor = (Tensor *)t;
  std::vector<size_t> sz = sizes(tensor);
  assert(sz[1] == 3);
  sz[0] = sz[0] * sz[1];
  sz[1] = 1;
  reshape(tensor, sz);
}

void backward_reshape(void *t) {
  auto *tensor = (Tensor *)t;
  std::vector<size_t> sz = sizes(tensor);
  assert(sz[0] % 3 == 0);
  sz[0] = sz[0] / 3;
  sz[1] = 3;
  reshape(tensor, sz);
}

void *sharpen(size_t &id, void *image) {
  void *gaussian = gaussianFilter();
  forward_reshape(image);
  void *blurred = wrapper_ConvLayer(
      std::to_string(id++).c_str(), image, gaussian, nullptr, 2, 2, 1, 1, 0, 0, -1, 0.0, 0.0);
  backward_reshape(image);
  backward_reshape(blurred);
  void *blurred_norm = normalize(id, blurred);
  void *image_norm = normalize(id, image);
  void *ret = wrapper_tensorMap2(
      std::to_string(id++).c_str(), (int)MathOp::AddWeighted, blurred_norm,
      image_norm);
  freeTensor(gaussian);
  freeTensor(blurred);
  freeTensor(blurred_norm);
  freeTensor(image_norm);
  return ret;
}

void *main_procedure(void *fg, void *bg) {
  size_t id = 1;
  void *g_bg = sharpen(id, bg);
  void *g_fg = sharpen(id, fg);
  void *ret = wrapper_tensorMap2(
      std::to_string(id++).c_str(), (int)MathOp::Blend2, g_bg, g_fg);
  std::cout << "**********************" << id << '\n';
  freeTensor(g_bg);
  freeTensor(g_fg);
  return ret;
}

extern void llvm_hpvm_initializeRuntimeController(
    const char *ConfigFile, const char *QRangeFile);

extern void llvm_hpvm_clearRuntimeController();

extern void llvm_hpvm_initTensorRt(int gpuid);

extern void llvm_hpvm_nextIter();

const size_t batch_size = 250;

int main(int argc, char *argv[]) {
  if (argc < 4)
    return 0;
  llvm_hpvm_initTensorRt(0);
  llvm_hpvm_initializeRuntimeController("data/tuner_confs.txt", "");

  startMemTracking();
  startProfiling();
  size_t bstart = 0;
  while (true) {
    auto *background = readDataSet(argv[1], bstart, batch_size, n_channels),
         *foreground = readDataSet(argv[2], bstart, batch_size, n_channels);
    if (!background || !foreground)
      break;

    auto *result = main_procedure(foreground, background);

    if (argc == 4) {
      saveDataSet(argv[3], (Tensor *)result, bstart);
      llvm_hpvm_imgInvokeRtControl(result, nullptr, bstart, bstart + batch_size);
    }
    else {
      saveDataSet(argv[3], (Tensor *)result, bstart, 10);
      auto *gold_output = readDataSet(argv[4], bstart, batch_size, n_channels);
      llvm_hpvm_imgInvokeRtControl(result, gold_output, bstart, bstart + batch_size);
    }

    bstart += batch_size;
    freeBatchMemory();
    clearTensorMap();
  }
  stopProfiling();
  llvm_hpvm_clearRuntimeController();
}
