#include "tensorUtils.h"
#include "tensor_runtime.h"
#include "visc.h"

const size_t n_channels = 3;

void *read_filter(const char *filter_prefix, const char *postfix, size_t sz) {
  void *t = readTrainedWeights(
      ((std::string)filter_prefix + postfix).c_str(), float_type,
      1, 1, sz, sz);
  auto *host_data = (float *)((Tensor *)t)->host_data;
  float sum = 0;
  for (size_t i = 0; i < sz * sz; i++)
    sum += host_data[i];
  sum = abs(sum);
  if (sum < 1e-4)
    return t;
  for (size_t i = 0; i < sz * sz; i++)
    host_data[i] /= sum;
  return t;
}

Tensor *gaussianFilter(float sigma, size_t w, size_t h, size_t n_chan) {
  int64_t m = (w - 1) / 2, n = (h - 1) / 2;
  auto *data = new float[w * h];
  float sum = 0.0f;
  for (int64_t i = -m; i <= m; i++)
    for (int64_t j = -n; j <= n; j++) {
      size_t idx = (i + m) * h + (j + n);
      float exponent = -(i * i + j * j) / (2.0 * sigma * sigma);
      data[idx] = exp(exponent);
      sum += data[idx];
    }
  if (sum != 0.0f)
    for (size_t i = 0; i < w * h; i++)
      data[i] /= sum;
  return (Tensor *)createFilterFromData(CUDNN_DATA_FLOAT, data, w, h, n_chan);
}

void *main_procedure(const char *filter_prefix, void *input) {
  // std::string prefix("../model_params/pipeline/filters/");
  // void *gaussian_filter = gaussianFilter(1.4, 3, 3, 1);
  void *gaussian_filter = read_filter(filter_prefix, "GaussianFilter.bin", 9);
  void *outline_filter = read_filter(filter_prefix, "OutlineFilter.bin", 3);
  void *sharpen_filter = read_filter(filter_prefix, "SharpenFilter.bin", 3);
  void *emboss_filter = read_filter(filter_prefix, "EmbossFilter.bin", 5);
  void *emboss_bias = read_filter(filter_prefix, "EmbossBias.bin", 1);
  void *motionblur_filter =
      read_filter(filter_prefix, "MotionblurFilter.bin", 9);

  // 0. Grayscale
  auto *summed_image = wrapper_tensorReduce("1", input, 1, (int)MathOp::Add);
  auto *grayscale_image =
      wrapper_tensorMap1("2", (int)MathOp::Avg3, summed_image);

  void *gaussian_out = wrapper_ConvLayer(
      "3", grayscale_image, gaussian_filter, nullptr, 4, 4, 1, 1, 0, 0, -1, 0.0,
      0.0);
  void *sharpen_out = wrapper_ConvLayer(
      "4", gaussian_out, sharpen_filter, nullptr, 1, 1, 1, 1, 0, 0, -1, 0.0,
      0.0);
  // void *outline_out = wrapper_ConvLayer(
  //   "7", grayscale_image, outline_filter, nullptr, 1, 1, 1, 1, 0, 0, -1, 0.0,
  //   0.0);
  void *motionblur_out = wrapper_ConvLayer(
      "5", sharpen_out, motionblur_filter, nullptr, 1, 1, 1, 1, 0, 0, -1, 0.0, 0.0);
  void *emboss_out = wrapper_ConvLayer(
      "6", motionblur_out, emboss_filter, nullptr, 1, 1, 1, 1, 0, 0, -1, 0.0, 0.0);
  void *emboss_bias_out =
      wrapper_tensorMap2("7", (int)MathOp::Add, emboss_out, emboss_bias);

  return emboss_bias_out;
}

const size_t batch_size = 100;

int main(int argc, char *argv[]) {
  if (argc < 4)
    return 0;
  llvm_hpvm_initTensorRt(0);
  llvm_hpvm_initializeRuntimeController("data/tuner_confs.txt", "");

  size_t bstart = 0;
  startMemTracking();
  startProfiling();
  while (true) {
    Tensor *batch = readDataSet(argv[2], bstart, batch_size);
    if (batch == nullptr)
      break;

    auto *result = main_procedure(argv[1], batch);
    if (argc == 4) {
      saveDataSet(argv[3], (Tensor *)result, bstart);
      llvm_hpvm_imgInvokeRtControl(
          result, nullptr, bstart, bstart + batch_size);
    } else {
      saveDataSet(argv[3], (Tensor *)result, bstart, 10);
      auto *gold_output = readDataSet(argv[4], bstart, batch_size, 1);
      llvm_hpvm_imgInvokeRtControl(
          result, gold_output, bstart, bstart + batch_size);
    }
    bstart += batch_size;
    freeBatchMemory();
    clearTensorMap();
  }
  stopProfiling();
  llvm_hpvm_clearRuntimeController();
}
